"""
The core Becca code

This package contains all the learning algorithms and heuristics
implemented together in a functional learning agent. Most users
won't need to modify anything in this package. Only core developers,
those who want to experiment with and improve the underlying
learning algorithms and their implementations, should need to muck
around with it.

See the becca_test package for examples of how to implement it with a world.
"""
